"""WAeUP data center.

The waeup data center cares for management of upload data and provides
tools for importing/exporting CSV data.
"""
import os
import shutil
import struct
import sys
import grok
from datetime import datetime
from zope.component import getUtility
from zope.interface import Interface
from waeup.interfaces import IWAeUPObject, IWAeUPCSVImporter, ICSVDataReceivers
from waeup.viewlets import Index, MainArea, LeftSidebar

class DataCenter(grok.Container):
    """A data center contains CSV files.
    """
    storage = os.path.join(os.path.dirname(__file__), 'files')

    def getReceivers(self):
        receivers = []
        curr_obj = getattr(self, '__parent__', None)
        while curr_obj is not None:
            if ICSVDataReceivers.providedBy(curr_obj):
                receivers = self.getCSVDataReceivers(curr_obj)
                break
            curr_obj = getattr(curr_obj, '__parent__', None)
        return receivers

    def getReceiverIds(self):
        """Get a dict of available receivers.

        The keys of the result are the receiver ids.
        """
        receivers = self.getReceivers()
        return dict([(self.getReceiverId(x), x) for x in receivers])
    
    def getImporters(self):
        """Get a list of all importers available.

        The search for available importers is done in two steps:

        1) Look for a utility providing ICSVDataReceiver,

        2) For every item of that utility: try to get an adapter
           providing IWAeUPCSVImporter.
        """
        result = []
        receivers = self.getReceivers()
        for receiver in receivers:
            try:
                importer = IWAeUPCSVImporter(receiver)
                result.append(importer)
            except:
                pass
        return result
    
    def getFiles(self):
        """Get a list of files stored in `storage`.
        """
        result = []
        if not os.path.exists(self.storage):
            return result
        for filename in os.listdir(self.storage):
            fullpath = os.path.join(self.storage, filename)
            if not os.path.isfile(fullpath):
                continue
            result.append(DataCenterFile(fullpath))
        return result

    def setStoragePath(self, path, move=False):
        """Set the path where to store files.
        """
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError('The path given does not exist: %s' % path)
        if move is True:
            # copy contents of old location to new one
            srcdir = self.storage
            dstdir = path
            for item in os.listdir(srcdir):
                if os.path.isdir(item):
                    shutil.copytree(os.path.join(srcdir, item),
                                    os.path.join(dstdir, item))
                else:
                    shutil.copy2(os.path.join(srcdir, item), dstdir)
        
        self.storage = path


    def getCSVDataReceivers(self, obj):
        """Get a list of attributes, that can receive CSV data.
        """
        result = []
        for attr_name in dir(obj):
            if attr_name.startswith('_'):
                continue
            try:
                attr = getattr(obj, attr_name)
                # This might fail...
                IWAeUPCSVImporter(attr)
                result.append(attr)
            except:
                pass
        return result

    def getPossibleImports(self):
        """Get list of possible imports.

        A single import is defined as a tuple
        
          ( <file-descr>, <importers> )
        
        where ``<file-descr>`` is an `IDataCenterFile` object and
        ``<importers>`` is a list of `IWAeUPCSVImporter` objects.
        """
        result = []
        importers = self.getImporters()
        for filedescr in self.getFiles():
            possible_importers = []
            for importer in importers:
                try:
                    # We could also try a complete parse...
                    importer.checkHeader(filedescr.context)
                except:
                    continue
                importer_context = (
                    importer, self.getReceiverId(importer.context))
                possible_importers.append(importer_context)
            if len(possible_importers) == 0:
                continue
            result.append((filedescr, possible_importers))
        return result

    def getReceiverId(self, obj):
        """Get a unique id for an object.

        If the object is stored in ZODB it should contain a `_p_oid`
        attribute, which is guaranteed to be unique over the ZODB.

        If the object has no such attribute, then it will be held in
        memory only and will vanish as soon as the request is over. In
        this case we can use the memory address of it.
        """
        if not hasattr(obj, '_p_oid'):
            return str(id(obj))
        oid = getattr(obj, '_p_oid')
        return str(struct.unpack('>Q', oid)[0])

        
class IDataCenterFile(Interface):
    """A data center file.
    """
    pass


class DataCenterFile(object):
    """A description of a file stored in data center.
    """
    grok.implements(IDataCenterFile)
    
    def __init__(self, context):
        self.context = context
        self.name = os.path.basename(self.context)
        self.size = self.getSize()
        self.uploaddate = self.getDate()

    def getDate(self):
        """Get a human readable datetime representation.
        """
        date = datetime.fromtimestamp(os.path.getctime(self.context))
        return date.strftime('%c')

    def getSize(self):
        """Get a human readable file size.
        """
        bytesize = os.path.getsize(self.context)
        size = "%s bytes" % bytesize
        units = ['kb', 'MB', 'GB']
        for power, unit in reversed(list(enumerate(units))):
            power += 1
            if bytesize >= 1024 ** power:
                size = "%.2f %s" % (bytesize/(1024.0**power), unit)
                break
        return size


class Content(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Index)

class Upload(grok.View):
    grok.context(DataCenter)
    grok.name('upload')
    grok.template('master')

    def update(self, uploadfile=None, CANCEL=None, SUBMIT=None):
        if CANCEL is not None:
            self.redirect(self.url(self.context))
            return
        if not uploadfile:
            return
        try:
            filename = uploadfile.filename
            target = os.path.join(self.context.storage, filename)
            open(target, 'wb').write(uploadfile.read())
        except IOError:
            self.flash('Error while uploading file. Please retry.')
            self.flash('I/O error: %s' % sys.exc_info()[1])
            return
        self.redirect(self.url(self.context))
    
class UploadMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Upload)
    

class Settings(grok.View):
    grok.context(DataCenter)
    grok.name('settings')
    grok.template('master')

    def update(self, newpath=None, save=None, cancel=None):
        if newpath is None:
            return
        if cancel is not None:
            self.redirect(self.url(self.context))
            return
        try:
            self.context.setStoragePath(newpath)
        except ValueError:
            self.flash('Given storage path cannot be used.')
            self.flash('Error: %s' %sys.exc_info()[1])
            return
        if newpath:
            self.redirect(self.url(self.context))
        return

class SettingsMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(Settings)

class SettingsLink(grok.Viewlet):
    grok.viewletmanager(LeftSidebar)
    grok.context(DataCenter)
    grok.view(Index)
    grok.order(2)
    grok.require('waeup.manageUniversity')

    def render(self):
        return u'<div class="portlet"><a href="@@settings">Settings</a></div>'

class ImportCSV(grok.View):
    """A view for importing things.

    """
    grok.require('waeup.manageUniversity')
    grok.template('master')

    def update(self, csvfile=None, clear=None, overwrite=None,
               receiverid=None, CANCEL=None, SUBMIT=None):
        if CANCEL is not None:
            self.redirect(self.url(self.context))
            return
        if not csvfile:
            return
        if not SUBMIT:
            return
        receivers = self.context.getReceiverIds()
        if receiverid not in receivers.keys():
            return
        receiver = receivers[receiverid]
        importer = IWAeUPCSVImporter(receiver)
        importer.doImport(csvfile)
        return

class Import(object):
    def __init__(self, filedescr, importers):
        self.file = filedescr
        self.importers = []
        for importer, receiverid in importers:
            importer.receiverid = receiverid
            self.importers.append(importer)
        
class ImportCSVMain(grok.Viewlet):
    grok.viewletmanager(MainArea)
    grok.context(DataCenter)
    grok.view(ImportCSV)
    grok.require('waeup.manageUniversity')

    def getImports(self):
        result = []
        imports = self.context.getPossibleImports()
        result = [Import(x, y) for x, y in imports]
        return result
    
    def update(self, *args, **kw):
        return super(ImportCSVMain, self).update(*args, **kw)
        
